searchData={"items":[{"type":"task","doc":"","title":"mix sorcery.init","ref":"Mix.Tasks.Sorcery.Init.html"},{"type":"function","doc":"","title":"Mix.Tasks.Sorcery.Init.run/1","ref":"Mix.Tasks.Sorcery.Init.html#run/1"},{"type":"module","doc":"To get started with Sorcery, let's use the generator\n\n`mix sorcery.init`\n\nBy default it creates a new namespace 'Src' in your app.\nThe src.ex module is very special, it starts up some :ets/:mnesia tables, as well as holds together some config for making everything else work.\n\nYou can manually create your own with \n```elixir\ndefmodule MyApp.Src do\n  use Sorcery\nend\n```\nAnd then make sure your queries and schemas, etc. are all namespaced below that like `MyApp.Src.Queries.GetStuff` and 'MyApp.Src.Schemas.Thing'","title":"Sorcery","ref":"Sorcery.html"},{"type":"behaviour","doc":"Functions for combining Sorcery and GenServer clients.\nPut these in your init function to create portals and receive updates.","title":"Sorcery.GenServerHelpers","ref":"Sorcery.GenServerHelpers.html"},{"type":"callback","doc":"Puts a :sorcery key in state, with a bunch of stuff used behind the scenes by Sorcery.\n\nThis is mandatory, before you can spawn any portals\n\nYou must pass in the sorcery_module. If you used the generator, it'll just be Src.","title":"Sorcery.GenServerHelpers.initialize_sorcery/2","ref":"Sorcery.GenServerHelpers.html#c:initialize_sorcery/2"},{"type":"callback","doc":"iex> body = %{sorcery_module: Src}\n    iex> state = initialize_sorcery(state, body)","title":"Examples - Sorcery.GenServerHelpers.initialize_sorcery/2","ref":"Sorcery.GenServerHelpers.html#c:initialize_sorcery/2-examples"},{"type":"callback","doc":"Get a list of entities at the given lvar. \n\nIf this lvar only exists in ONE portal, you can leave out the portal_name, just know that it might be imperceptibly slower since the function needs to iterate over all portals until it finds one with that lvar..\n\n```elixir\n<% players = portal_view(@sorcery, \"?all_players\") %>\n```","title":"Sorcery.GenServerHelpers.portal_view/2","ref":"Sorcery.GenServerHelpers.html#c:portal_view/2"},{"type":"callback","doc":"Get a list of entities at the given portal/lvar\nYou probably want to use this inside heex\n```elixir\n<% players = portal_view(@sorcery, :my_portal, \"?all_players\") %>\n```","title":"Sorcery.GenServerHelpers.portal_view/3","ref":"Sorcery.GenServerHelpers.html#c:portal_view/3"},{"type":"callback","doc":"Creates a portal between the GenServer and the PortalServer, using the given arguments.\n\nTakes the state, returns it unchanged. An update should happen in a few milliseconds to add more data to the state..","title":"Sorcery.GenServerHelpers.spawn_portal/2","ref":"Sorcery.GenServerHelpers.html#c:spawn_portal/2"},{"type":"callback","doc":"iex> body = %{portal_server: Postgres, portal_name: :my_portal, query_module: MyQuery, query_args: %{player_id: 1}}\n    iex> state = spawn_portal(state, body)","title":"Examples - Sorcery.GenServerHelpers.spawn_portal/2","ref":"Sorcery.GenServerHelpers.html#c:spawn_portal/2-examples"},{"type":"module","doc":"This is a garbage collector, of sorts.\n\nEvery time it runs, it will scan SorceryDb for all dead pids and hang onto those until the postmortem_delay is over\nFor every pid in which the delay IS over, it will find the entities in their portals, filter out any entities that are being watched in other portals, and then remove them from :mnesia.\n```elixir\ndefmodule MyApp.Sorcery do\n  use Sorcery,\n    killbot: %{\n      # Defaults to 10 minutes\n      interval: 10 * 60 * 1_000, \n\n      # Wait 2 intervals AFTER the pid dies, before wiping the entities       \n      # It is likely a user will close a page and come back a few minutes later...\n      # So we don't want to uncache things too quickly.\n      #\n      # Again, I stress that this is NOT milliseconds, but the number of intervals.\n      postmortem_delay: 2        \n    }\nend\n```","title":"Sorcery.Killbot","ref":"Sorcery.Killbot.html"},{"type":"function","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","title":"Sorcery.Killbot.child_spec/1","ref":"Sorcery.Killbot.html#child_spec/1"},{"type":"function","doc":"Return all the dead portals being watched, which have not yet been removed.","title":"Sorcery.Killbot.get_watched/0","ref":"Sorcery.Killbot.html#get_watched/0"},{"type":"function","doc":"Start a run immediately. This does speed up the postmortem_delay","title":"Sorcery.Killbot.run_now/0","ref":"Sorcery.Killbot.html#run_now/0"},{"type":"behaviour","doc":"Functions for combining Sorcery and LiveViews","title":"Sorcery.LiveHelpers","ref":"Sorcery.LiveHelpers.html"},{"type":"callback","doc":"Puts a :sorcery key in socket.assigns, with a bunch of stuff used behind the scenes by Sorcery.\n\nThis is mandatory, before you can spawn any portals\n\nYou must pass in the sorcery_module. If you used the generator, it'll just be Src.","title":"Sorcery.LiveHelpers.initialize_sorcery/2","ref":"Sorcery.LiveHelpers.html#c:initialize_sorcery/2"},{"type":"callback","doc":"iex> body = %{sorcery_module: Src}\n    iex> socket = initialize_sorcery(socket, body)","title":"Examples - Sorcery.LiveHelpers.initialize_sorcery/2","ref":"Sorcery.LiveHelpers.html#c:initialize_sorcery/2-examples"},{"type":"callback","doc":"Get a list of entities at the given lvar. \n\nIf this lvar only exists in ONE portal, you can leave out the portal_name, just know that it might be imperceptibly slower since the function needs to iterate over all portals until it finds one with that lvar..\n\n```elixir\n<% players = portal_view(@sorcery, \"?all_players\") %>\n```","title":"Sorcery.LiveHelpers.portal_view/2","ref":"Sorcery.LiveHelpers.html#c:portal_view/2"},{"type":"callback","doc":"Get a list of entities at the given portal/lvar\nYou probably want to use this inside heex\n```elixir\n<% players = portal_view(@sorcery, :my_portal, \"?all_players\") %>\n```","title":"Sorcery.LiveHelpers.portal_view/3","ref":"Sorcery.LiveHelpers.html#c:portal_view/3"},{"type":"callback","doc":"Creates a portal between the LiveView and the PortalServer, using the given arguments.\n\nTakes the socket, returns it unchanged.","title":"Sorcery.LiveHelpers.spawn_portal/2","ref":"Sorcery.LiveHelpers.html#c:spawn_portal/2"},{"type":"callback","doc":"iex> body = %{portal_server: Postgres, portal_name: :my_portal, query_module: MyQuery, query_args: %{player_id: 1}}\n    iex> socket = spawn_portal(socket, body)","title":"Examples - Sorcery.LiveHelpers.spawn_portal/2","ref":"Sorcery.LiveHelpers.html#c:spawn_portal/2-examples"},{"type":"module","doc":"This is the interface for creating changes in Sorcery.\nMutations have some similarities to Ecto Changesets in the sense that we are building a struct that defines all the changes without actually applying them.\nBut there are many differences, as you will soon see.\n\nYou start by creating the initial mutation with a portal.\n\nThen you apply all the changes you want to the data in that portal.\n\nAt first nothing happens! This is because you must send the mutation up to the parent PortalServer.\n\nAfter it applies the changes, it will dispatch them back down to any child PortalServers that care about any of the entities involved.\nThis is where the magic of Sorcery really comes in. Under the hood, we are doing some pretty efficient reverse queries, in parallel, to figure out where to send the data.\n\nAll the PortalServers will update their Portals accordingly.\n\n```elixir\niex> alias Sorcery.Mutation, as: M\niex> m = M.init(socket.assigns.sorcery, :my_portal)\niex> m = M.update(m, [:player, 1, :age], fn _original_age, latest_age -> latest_age + 1 end)\niex> m = M.put(m, [:player, 1, :health], 100)\niex>","title":"Sorcery.Mutation","ref":"Sorcery.Mutation.html"},{"type":"module","doc":"iex> m = M.create_entity(m, :team, \"?my_new_team\", %{name: \"My New Team\"})\niex> m = M.put(m, [:player, 1, :team_id], \"?my_new_team.id\")\niex> # This funny string is of course, the placeholder. But after we actually run the mutation, the team is created with a normal integer id.\niex> # The parent PortalServer creates the new team entity, it will automatically replace all calls to \"?my_new_team\" with that entity.\niex> Sorcery.Mutation.send_mutation(m)\niex> # ... a few milliseconds later, after receiving the update\niex> player = portal_view(@sorcery, :my_portal, \"?all_players\")[1]\niex> new_team_id = player.team_id\niex> is_integer(new_team_id)\ntrue\niex> team = portal_view(@sorcery, :my_portal, \"?all_teams\")[new_team_id]\niex> team.name\n\"My New Team\"\n```","title":"We can also use placeholder ids so that new entities (which haven't even been created yet) can be referenced by other entities. - Sorcery.Mutation","ref":"Sorcery.Mutation.html#module-we-can-also-use-placeholder-ids-so-that-new-entities-which-haven-t-even-been-created-yet-can-be-referenced-by-other-entities"},{"type":"function","doc":"One of the benefits of a Mutation is that you can create entities using placeholder ids.","title":"Sorcery.Mutation.create_entity/4","ref":"Sorcery.Mutation.html#create_entity/4"},{"type":"function","doc":"iex> m = M.create_entity(m, :team, \"?my_new_team\", %{name: \"My New Team\"})\n    iex> m = M.put(m, [:player, 1, :team_id], \"?my_new_team\")\n    iex> # It always defaults to using the :id, but you can specify another field\n    iex> m = M.put(m, [:player, 1, :team_id], \"?my_new_team.my_field\")","title":"Examples - Sorcery.Mutation.create_entity/4","ref":"Sorcery.Mutation.html#create_entity/4-examples"},{"type":"function","doc":"Sometimes we want to delete an entity entirely. Be careful, after the mutation is run, this cannot be undone","title":"Sorcery.Mutation.delete_entity/3","ref":"Sorcery.Mutation.html#delete_entity/3"},{"type":"function","doc":"The first step of every mutation. You must have an existing portal in order to do any of this.","title":"Sorcery.Mutation.init/2","ref":"Sorcery.Mutation.html#init/2"},{"type":"function","doc":"iex> m = Sorcery.Mutation.init(socket.assigns.sorcery, :my_portal)\n    iex> is_struct(m)\n    true","title":"Examples - Sorcery.Mutation.init/2","ref":"Sorcery.Mutation.html#init/2-examples"},{"type":"function","doc":"Just like put_in, but for mutations.","title":"Sorcery.Mutation.put/3","ref":"Sorcery.Mutation.html#put/3"},{"type":"function","doc":"iex> m = Sorcery.Mutation.put(m, [:player, 1, :health], 100)","title":"Examples - Sorcery.Mutation.put/3","ref":"Sorcery.Mutation.html#put/3-examples"},{"type":"function","doc":"Sends the mutation to the corresponding Portal Server\nThe Portal Server will then update its own data store.\nAll portals that care about the changes will automatically update accordingly (including the one calling this function) \n\nReturns the mutation passed in.\n\nTo be clear, this does not return the new data you are waiting for. You probably won't need the return value.\nAll the updates happen automatically, through some ~~magic~~  sorcery behind the scenes.","title":"Sorcery.Mutation.send_mutation/2","ref":"Sorcery.Mutation.html#send_mutation/2"},{"type":"function","doc":"Kind of like a fusion between Kernel.update_in/3 and Map.update/4\n\nYou need to pass in both a path list, and a default value in case the original value is nil.","title":"Sorcery.Mutation.update/3","ref":"Sorcery.Mutation.html#update/3"},{"type":"function","doc":"iex> Sorcery.Mutation.update(m, [:player, 1, :age], 100, fn _original_age, latest_age -> latest_age + 1 end)","title":"Examples - Sorcery.Mutation.update/3","ref":"Sorcery.Mutation.html#update/3-examples"},{"type":"module","doc":"","title":"Sorcery.Mutation.Temp","ref":"Sorcery.Mutation.Temp.html"},{"type":"function","doc":"","title":"Sorcery.Mutation.Temp.add_temp_portal/2","ref":"Sorcery.Mutation.Temp.html#add_temp_portal/2"},{"type":"module","doc":"A PortalServer is a special type of GenServer. Portals are created between different PortalServers, always in a hierarchical parent <--> child relationship.\n\nEvery PortalServer must have access to some kind of data store. That might be Postgres via Ecto.Repo... or it could be a LiveView storing data in socket.assigns. (Remember, LiveViews *are* GenServers!)\nYou can even create your own adapter and use ANY backend as the data store, as long as you can query and mutate the data.","title":"Sorcery.PortalServer","ref":"Sorcery.PortalServer.html"},{"type":"module","doc":"A PortalServer is any GenServer that holds a special :sorcery key somewhere in its state, and also handles messages sent to {:sorcery, msg}","title":"Setup - Sorcery.PortalServer","ref":"Sorcery.PortalServer.html#module-setup"},{"type":"function","doc":"To convert any process into a PortalServer, you must follow 2 steps:\n1) Add some PortalServer configuration to the state\n2) Implement the PortalServer handlers for a handle_info function\n\n\n```elixir\nuse GenServer\nuse Sorcery.PortalServer\n\ndef init(_) do\n  state = %{} # You can still add whatever you want here\n\n  state = Sorcery.PortalServer.add_portal_server_state(state, %{\n    config_module: Src,\n    store_adapter: Sorcery.StoreAdapters.Ecto,\n\n    # This depends on the adapters you use\n    args: %{\n      repo_module: MyApp.Repo\n    }\n  })\n\n  {:ok, state}\nend\n```","title":"Sorcery.PortalServer.add_portal_server_state/2","ref":"Sorcery.PortalServer.html#add_portal_server_state/2"},{"type":"module","doc":"This exists to resolve a lot of confusion caused by generically using the word 'state' to refer to either the data inside the :sorcery key, or the data *including* :sorcery.\nFor example a LiveView would have socket.assigns.sorcery\nSo does state refer to socket.assigns? Or socket.assigns.sorcery?\nIn theory, Sorcery should never need to access the outer map, aside from maybe the helpers at most.\n\nSo to save confusion, InnerState is ONLY ever the value of the :sorcery key.\nThere is no OuterState struct, as it depends on how the user configures the PortalServer.","title":"Sorcery.PortalServer.InnerState","ref":"Sorcery.PortalServer.InnerState.html"},{"type":"function","doc":"","title":"Sorcery.PortalServer.InnerState.new/1","ref":"Sorcery.PortalServer.InnerState.html#new/1"},{"type":"module","doc":"A Portal represents an ongoing SrcQL Query.\nRather than grabbing data and ending, it continues watching for changes to the results set.\n\nA copy of any given Portal is found in the state of two different PortalServers; the parent, and the child PortalServer.","title":"Sorcery.PortalServer.Portal","ref":"Sorcery.PortalServer.Portal.html"},{"type":"function","doc":"Get the list of all lvars that match a given tk","title":"Sorcery.PortalServer.Portal.all_lvars_by_tk/2","ref":"Sorcery.PortalServer.Portal.html#all_lvars_by_tk/2"},{"type":"function","doc":"When we freeze a portal, all the lvars are replaced with tablekeys (tk).\nIf multiple lvars reference the same tablekey, they are merged.\n\nA frozen portal is no longer kept up to date. Instead, it is immutable.","title":"Sorcery.PortalServer.Portal.freeze/1","ref":"Sorcery.PortalServer.Portal.html#freeze/1"},{"type":"function","doc":"","title":"Sorcery.PortalServer.Portal.get_in/3","ref":"Sorcery.PortalServer.Portal.html#get_in/3"},{"type":"function","doc":"","title":"Sorcery.PortalServer.Portal.new/1","ref":"Sorcery.PortalServer.Portal.html#new/1"},{"type":"behaviour","doc":"A query module defines, in plain elixir data structures, the kind of data we want to watch for.\n\nThe syntax takes some inspiration from Datalog, but with many differences as well.\n\n```elixir\ndefmodule Src.Queries.GetBattle do\n\n  use Sorcery.Query, %{\n\n    # Args will be passed in later when the query is called.\n    args: %{\n      player_id: :integer\n    },\n\n    # This is the meat of any query. Read it one row at a time.\n    # Every row has either 3, 4, columns\n    where: [\n      # 4 column syntax:\n      # [lvar,        tk,            attr,         value]\n      #\n      #\n      # So we start with an lvar (or 'Logic Variable') called \"?player\"\n      # It represents a set of entities with the Schema of :player\n      # And we are filtering them such that ?player.id == args.player_id\n      # The arg MUST be declared in the args map.\n      [ \"?player\",    :player,       :id,          :args_player_id],\n\n      # Now we make a new lvar called \"?team\"\n      # See how it now references the previous lvar?\n      # So we are filtering all teams such that team.id matches ANY ?player.team_id\n      [ \"?team\",      :team,         :id,          \"?player.team_id\"],\n      [ \"?arena\",     :battle_arena, :id,          \"?team.location_id\"],\n\n      # This is not the same as ?team.\n      # Its a new lvar using the same schema, but with a different set of filters\n      # So we're getting all teams such that team.location_id == ?arena.id\n      [ \"?all_teams\", :team,         :location_id, \"?arena.id\"],\n\n      # Now we use the 3 column syntax, just to avoid repetition.\n      # This could also be rewritten as two rows:\n      # [\"?all_players\", :player, :team_id, \"?all_teams.id\"],\n      # [\"?all_players\", :player, :health, {:>, 0}],\n      [ \"?all_players\", :player, [\n        {:team_id, \"?all_teams.id\"},\n        {:health, {:>, 0}},\n      ]],\n      # Notice the value above {:>, 0}\n      # By default, every value automatically expands under the hood to {:==, value}\n      # But if you want to manually use an operator, you can.\n      # Possible operators: :==, :!=, :>, :>=, :<, :<=, :in\n\n\n      [ \"?spells\", :spell_instance, :player_id, \"?all_players.id\"],\n      [ \"?spell_types\", :spell_type, :id, \"?spells.type_id\"],\n    ],\n\n    # Without a find map, the query returns no results.\n    # We do not necessarily need all the lvars, nor all the fields\n    # If we want all available fields, use :*\n    # Otherwise pass in a list of specific ones. The :id attr is automatically added.\n    find: %{\n      \"?arena\" => :*,\n      \"?all_teams\" => [:name, :location_id],\n      \"?all_players\" => :*,\n      \"?spells\" => :*,\n      \"?spell_types\" => :*,\n    }\n  }\n\nend\n\n```","title":"Sorcery.Query","ref":"Sorcery.Query.html"},{"type":"callback","doc":"Return a list of WhereClause structs for the Query module","title":"Sorcery.Query.clauses/1","ref":"Sorcery.Query.html#c:clauses/1"},{"type":"callback","doc":"iex> [clause1 | _] = Src.Queries.GetBattle.clauses(%{player_id: 1})\n    iex> clause1\n    %Sorcery.Query.WhereClause{lvar: :\"?player\", tk: :player, attr: :id, left: nil, right: 1, op: :==, other_lvar: nil, other_lvar_attr: nil, arg_name: nil, right_type: :literal}","title":"Examples - Sorcery.Query.clauses/1","ref":"Sorcery.Query.html#c:clauses/1-examples"},{"type":"callback","doc":"Returns the finds. This differs from raw_struct().finds because :id fields have been added.","title":"Sorcery.Query.finds/0","ref":"Sorcery.Query.html#c:finds/0"},{"type":"callback","doc":"iex> Src.Queries.GetBattle.finds()\n    %{\"?arena\": :*, \"?all_teams\": [:location_id, :id, :name], \"?all_players\": :*, \"?spells\": :*, \"?spell_types\": :*}","title":"Examples - Sorcery.Query.finds/0","ref":"Sorcery.Query.html#c:finds/0-examples"},{"type":"function","doc":"If you have a map in the format of `%{tk => %{id => %{...entity...}}}`\nThen you can use it like a database and run the query against it.","title":"Sorcery.Query.from_tk_map/3","ref":"Sorcery.Query.html#from_tk_map/3"},{"type":"callback","doc":"Returns a Sorcery.Query struct.","title":"Sorcery.Query.raw_struct/0","ref":"Sorcery.Query.html#c:raw_struct/0"},{"type":"callback","doc":"iex> q = Src.Queries.GetBattle.raw_struct()\n    iex> is_struct(q, Sorcery.Query)\n    iex> q.find[\"?all_players\"]\n    :*\n    iex> [clause1 | _] = q.where\n    iex> clause1\n    [\"?player\", :player, :id, :args_player_id]\n    iex> Enum.at(q.lvar_tks, 3)\n    {\"?all_teams\", :team}","title":"Examples - Sorcery.Query.raw_struct/0","ref":"Sorcery.Query.html#c:raw_struct/0-examples"},{"type":"callback","doc":"A unique list of all tks mentioned by this query.","title":"Sorcery.Query.tks_affected/0","ref":"Sorcery.Query.html#c:tks_affected/0"},{"type":"callback","doc":"iex> Src.Queries.GetBattle.tks_affected()\n    [:player, :team, :battle_arena, :spell_instance, :spell_type]","title":"Examples - Sorcery.Query.tks_affected/0","ref":"Sorcery.Query.html#c:tks_affected/0-examples"},{"type":"type","doc":"","title":"Sorcery.Query.t/0","ref":"Sorcery.Query.html#t:t/0"},{"type":"module","doc":"This is the return type of any forward query, and any generated data.\n\nTry not to access the data directly, but rather use the functions in this module.\n\nThe reason is that Sorcery is still a work in progress, the format might change, but the functions should never break their contract's.\n\nThis is also the basic format of portals.\n\nCurrently the :primary_entities field is not being used. The idea was for it to help with ergonomics when a query is only expected to have one result.","title":"Sorcery.ReturnedEntities","ref":"Sorcery.ReturnedEntities.html"},{"type":"function","doc":"Takes the :find key from some SrcQL and uses that to prune the RE.","title":"Sorcery.ReturnedEntities.apply_find_map/2","ref":"Sorcery.ReturnedEntities.html#apply_find_map/2"},{"type":"function","doc":"iex> re = Sorcery.ReturnedEntities.new()\n    iex> re = put_entities(re, :bird, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> re = put_entities(re, :spell, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> get_entities(re, :bird)\n    [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"} ]\n    iex> re = apply_find_map(re, %{bird: [:id]})\n    iex> get_entities(re, :bird)\n    [ %{id: 1}, %{id: 23} ]","title":"Examples - Sorcery.ReturnedEntities.apply_find_map/2","ref":"Sorcery.ReturnedEntities.html#apply_find_map/2-examples"},{"type":"function","doc":"","title":"Sorcery.ReturnedEntities.assign_lvar_tk/3","ref":"Sorcery.ReturnedEntities.html#assign_lvar_tk/3"},{"type":"function","doc":"Deletes attribute keys from every entity of the given tk","title":"Sorcery.ReturnedEntities.delete_attrs/3","ref":"Sorcery.ReturnedEntities.html#delete_attrs/3"},{"type":"function","doc":"iex> re = Sorcery.ReturnedEntities.new()\n    iex> re = put_entities(re, :player, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> re.data.player\n    %{1 => %{id: 1, name: \"A\"}, 23 => %{id: 23, name: \"B\"}}\n    iex> re = delete_attrs(re, :player, [ :name ])\n    iex> re.data.player\n    %{1 => %{id: 1}, 23 => %{id: 23}}","title":"Examples - Sorcery.ReturnedEntities.delete_attrs/3","ref":"Sorcery.ReturnedEntities.html#delete_attrs/3-examples"},{"type":"function","doc":"Removes all the listed ids of a given tk.","title":"Sorcery.ReturnedEntities.delete_entities/3","ref":"Sorcery.ReturnedEntities.html#delete_entities/3"},{"type":"function","doc":"iex> re = Sorcery.ReturnedEntities.new()\n    iex> re = put_entities(re, :player, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> re.data.player\n    %{1 => %{id: 1, name: \"A\"}, 23 => %{id: 23, name: \"B\"}}\n    iex> re = delete_entities(re, :player, [ 23 ])\n    iex> re.data.player\n    %{1 => %{id: 1, name: \"A\"}}","title":"Examples - Sorcery.ReturnedEntities.delete_entities/3","ref":"Sorcery.ReturnedEntities.html#delete_entities/3-examples"},{"type":"function","doc":"","title":"Sorcery.ReturnedEntities.get_entities/2","ref":"Sorcery.ReturnedEntities.html#get_entities/2"},{"type":"function","doc":"iex> re = Sorcery.ReturnedEntities.new()\n    iex> re = put_entities(re, :player, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> get_entities(re, :player)\n    [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"} ]","title":"Examples - Sorcery.ReturnedEntities.get_entities/2","ref":"Sorcery.ReturnedEntities.html#get_entities/2-examples"},{"type":"function","doc":"Creates a new, empty RE struct.","title":"Sorcery.ReturnedEntities.new/0","ref":"Sorcery.ReturnedEntities.html#new/0"},{"type":"function","doc":"iex> new()\n    %Sorcery.ReturnedEntities{ primary_entities: [], data: %{} }","title":"Examples - Sorcery.ReturnedEntities.new/0","ref":"Sorcery.ReturnedEntities.html#new/0-examples"},{"type":"function","doc":"","title":"Sorcery.ReturnedEntities.put_entities/3","ref":"Sorcery.ReturnedEntities.html#put_entities/3"},{"type":"function","doc":"iex> re = Sorcery.ReturnedEntities.new()\n    iex> re = put_entities(re, :player, [ %{id: 1, name: \"A\"}, %{id: 23, name: \"B\"}])\n    iex> re.data.player\n    %{1 => %{id: 1, name: \"A\"}, 23 => %{id: 23, name: \"B\"}}\n    iex> re = put_entities(re, :player, [ %{id: 1, name: \"C\"}, %{id: 42, name: \"B\"}])\n    iex> re.data.player\n    %{1 => %{id: 1, name: \"C\"}, 23 => %{id: 23, name: \"B\"}, 42 => %{id: 42, name: \"B\"}}","title":"Examples - Sorcery.ReturnedEntities.put_entities/3","ref":"Sorcery.ReturnedEntities.html#put_entities/3-examples"},{"type":"module","doc":"In sorcery, you must define your entity types as 'Schemas'\n\n\n```elixir\ndefmodule MyApp.Player do\n  use Sorcery.Schema, fields: %{\n    name: %{t: :string, min: 0, max: 45, default: \"Nameless\"},\n    age: %{t: :integer, min: 0, max: 99, optional?: false},\n  }\nend\n```\n\nSee the guide for a comprehensive table of all possible attributes for each :t.\n\nAfter you do the above setup, The MyApp.Player module will have some neat superpowers. Go take a look in iex.","title":"Sorcery.Schema","ref":"Sorcery.Schema.html"},{"type":"function","doc":"","title":"Sorcery.Schema.meta_defaults/1","ref":"Sorcery.Schema.html#meta_defaults/1"},{"type":"behaviour","doc":"WIP\ncurrently it is not possible to add your own custom field types. But in the future it should be possible since they all follow this behaviour","title":"Sorcery.Schema.FieldType","ref":"Sorcery.Schema.FieldType.html"},{"type":"callback","doc":"","title":"Sorcery.Schema.FieldType.get_sd_field/1","ref":"Sorcery.Schema.FieldType.html#c:get_sd_field/1"},{"type":"callback","doc":"During validation of a field, we go through every attribute, comparing it to the given value to see if everything is valid.\n\nWe do it this way to try to keep everything related to attributes all in one place - the field struct.\nFor example, if you look in Sorcery.Schema.FieldType.Integer, you will find a defstruct that lists several attributes. \nBut there is also the is_valid? function, with a different head for each one.","title":"Sorcery.Schema.FieldType.is_valid?/3","ref":"Sorcery.Schema.FieldType.html#c:is_valid?/3"},{"type":"function","doc":"","title":"Sorcery.Schema.FieldType.new/1","ref":"Sorcery.Schema.FieldType.html#new/1"},{"type":"callback","doc":"","title":"Sorcery.Schema.FieldType.new/1","ref":"Sorcery.Schema.FieldType.html#c:new/1"},{"type":"function","doc":"","title":"Sorcery.Schema.FieldType.new/2","ref":"Sorcery.Schema.FieldType.html#new/2"},{"type":"behaviour","doc":"A store adapter is a module that allows a PortalServer to access a data store by taking SrcQL, and converting it into something compatible with the specific store.\n\nCurrently there are only two Adapters available: Ecto, and InMemory.\nInMemory is used behind the scenes by the LiveHelper, and is basically a noop, preferring to keep all data in the portal itself.\nWhile the `Sorcery.StoreAdapter.Ecto` adapter is used for dealing with serious backends like MySql and Postgres.","title":"Sorcery.StoreAdapter","ref":"Sorcery.StoreAdapter.html"},{"type":"function","doc":"","title":"Sorcery.StoreAdapter.mutation/3","ref":"Sorcery.StoreAdapter.html#mutation/3"},{"type":"function","doc":"","title":"Sorcery.StoreAdapter.query/4","ref":"Sorcery.StoreAdapter.html#query/4"},{"type":"callback","doc":"","title":"Sorcery.StoreAdapter.run_mutation/2","ref":"Sorcery.StoreAdapter.html#c:run_mutation/2"},{"type":"callback","doc":"","title":"Sorcery.StoreAdapter.run_query/3","ref":"Sorcery.StoreAdapter.html#c:run_query/3"},{"type":"module","doc":"This adapter requires the repo module to be passed in as args. \n\nHere is an example PortalServer\n\n```elixir\ndefmodule Src.PortalServers.Postgres do\n  use GenServer\n\n  def init(_) do\n    state = %{} # You can still add whatever you want here\n\n    state = Sorcery.PortalServer.add_portal_server_state(state, %{\n      config_module: Src,  # See below\n      store_adapter: Sorcery.StoreAdapter.Ecto,\n\n      args: %{\n        repo_module: MyApp.Repo\n      }\n    })\n    {:ok, state}\n  end\n\n\n  def handle_info({:sorcery, msg}, %{sorcery: inner_state} = outer_state) do\n    inner_state = Sorcery.PortalServer.handle_info(msg, inner_state)\n    {:noreply, Map.put(outer_state, :sorcery, inner_state)}\n  end\n\n\nend\n```","title":"Sorcery.StoreAdapter.Ecto","ref":"Sorcery.StoreAdapter.Ecto.html"},{"type":"extras","doc":"# Sorcery\n\nA framework which rethinks how data flows, and how we build apps.\n\nPlays nicely with Phoenix LiveViews, or can be used alone.\n\nThe philosophy of Sorcery:\n\n```\nA database has a table called People.\n\nThere are 3 different web pages that all display *some* data about the row where People.id == 1\nNo matter where this person appears, they are always the same entity.\n\nIf they change their name in one place, it should automatically update their name in EVERY place that it appears.\n\nThis should be the DEFAULT way of building an app. \nI don't want to setup a Phoenix.Channel, nor a PubSub.\nI don't have time to wire this up in every page where I want it to happen.\nIt should be automatic.\n\nWe should not even need to think about it. I use an entity in a page, and it should stay up to date forever. Period.\n```","title":"Sorcery","ref":"readme.html"},{"type":"extras","doc":"These docs are horribly incomplete. But you can clone and run the demo app [https://github.com/greetingsfellowhumans/sorcery_demo.git](https://github.com/greetingsfellowhumans/sorcery_demo.git)\nIt is extremely minimalistic. No css whatsoever... but it shows how everything works together.","title":"See the demo - Sorcery","ref":"readme.html#see-the-demo"},{"type":"extras","doc":"Sorcery comes with it's own headless, reversible, query language (SrcQL). By default it uses Ecto and whatever backend you want (Postgres, etc.). \nYou run the query and get a 'Portal' which is like a little wormhole showing whatever database entities match the Query.\n\nHere we see it in LiveViews, but any GenServer can do this.\nThe query lists several 'Logic Variables' or 'Lvars', and we are looking at one called \"?all_players\"\n\n```elixir\n~H\"\"\"\n  <%= for %{id: id, health: health, name: name} = _player <- portal_view(@sorcery, :my_portal, \"?all_players\") do %>\n     <%= id %> | <%= name %>'s health: <%= health %> \n  <% end %>\n\"\"\"\n```\n\nIf someone new enters the room, you'll see the change instantly. Ditto if someone leaves, or changes their name, or anything like that.\n\n\nSo a Portal is almost like a PubSub, if a PubSub had access to a full featured language, and the ability to recursively watch OTHER PubSubs.","title":"But how *does* it work?! - Sorcery","ref":"readme.html#but-how-does-it-work"},{"type":"extras","doc":"Memory hog with almost no latency.\n\nAll entities being watched are cached in mnesia tables. So often this allows us to skip calling the database, and use cache.","title":"Performance - Sorcery","ref":"readme.html#performance"},{"type":"extras","doc":"```elixir\ndef deps do\n  [\n    {:sorcery, \"~> 0.3.0\"},\n  ]\nend\n```\n\nThen run the following to get bootstrapped\n```bash\n$ mix deps.get\n$ mix sorcery.init\n```\n\nSeveral files were just created in your app under /lib/src\n\nMake sure to also open your `application.ex` and add\n```elixir\nchildren = [\n  # Just an example, but you need to add each of your PortalServers here.\n  {Src.PortalServers.Postgres, name: Src.PortalServers.Postgres},\n\n  # This is mandatory.\n  {Src, []},\n]\n```\n\nThe next thing you should do is create some schemas, and queries.","title":"Installation - Sorcery","ref":"readme.html#installation"},{"type":"extras","doc":"# Changelog\n## [0.3.5] - 2024-06-18","title":"Changelog","ref":"changelog.html"},{"type":"extras","doc":"Mutation.create_entity was broken. I fixed it.\n\n## [0.3.4] - 2024-06-17","title":"Fixed - Changelog","ref":"changelog.html#fixed"},{"type":"extras","doc":"- New field types for schemas: float, boolean, list, map\nNote that list fields must contain a :coll_of type. for example:\n```elixir\nmy_items: %{t: :list, coll_of: :integer}\n```\n\nWhile using ecto/postgres, lists and maps are stored as json. I have not tested this with anything else\n\n## [0.3.3] - 2024-06-17","title":"Added - Changelog","ref":"changelog.html#added"},{"type":"extras","doc":"- The Killbot module\n  This is basically a garbage collection system. Since SorceryDb stores a lot of entities in :mnesia, and some other portal data in :ets, Killbot is a process that periodically checks which entities are no longer being used anywhere, and safely removes them.","title":"Added - Changelog","ref":"changelog.html#added"},{"type":"extras","doc":"New options for your Src module\n```elixir\nuse Sorcery,\n    killbot: %{\n      interval: int,\n      postmortem_delay: int\n    }\n```\nThese are optional. \nIn theory, Killbot should not require any setup, it is initialized by the `use Sorcery` under the hood.","title":"Changed - Changelog","ref":"changelog.html#changed"},{"type":"extras","doc":"None (in theory)\n\n## [Unreleased]\n- Most likely optimistic updates will break when the mutation creates new entities. Still need to figure out what to do there.\n- Create the killbot module to periodically trim data from SorceryDb when pids are dead.\n- Better document SorceryDb.\n\n## [0.3.2] - 2024-06-01","title":"Breaking Changes - Changelog","ref":"changelog.html#breaking-changes"},{"type":"extras","doc":"- Optimistic Updates\n  Now when you send a mutation, it will return a :temp_data field, guessing at the changes. This allows us to see the updated data long before the changes reach the parent portal_server or database.\n  temp_data is reset after the real data comes in\n  Mutations cannot be run on a portal that already has temp_data. That would be too complex. Extra mutations are silently dropped.\n- More documentation, especially for Sorcery.Query","title":"Added - Changelog","ref":"changelog.html#added"},{"type":"extras","doc":"- %Sorcery.PortalServer.InnerState{}\n  This has eased a lot of confusion and unveiled bugs waiting to happen.\n- Turned Sorcery.Query into a behaviour, and the __using__ macro into an implementation of it.\n    Now in exdoc, all the functions defined by Use are well documented with examples that pass tests.","title":"Changed - Changelog","ref":"changelog.html#changed"},{"type":"extras","doc":"","title":"live_views","ref":"live_views.html"},{"type":"extras","doc":"Let's start with the minimalistic demo module for a LiveView, and explain it afterward\n\n```elixir\ndefmodule MyAppWeb.SandboxLive do\n  use Phoenix.LiveView\n  use Sorcery.LiveHelpers  # <- This adds functions like spawn_portal/2 etc.\n\n  def mount(_param, _sesh, socket) do\n    socket = \n      socket\n      |> initialize_sorcery(%{\n        sorcery_module: MyApp.Sorcery # We talk about this module in the Introduction guide\n      })\n\n      # You can call this multiple times with different config\n      # As long as the portal_name is unique on a per-liveview basis.\n      |> spawn_portal(%{\n        portal_name: :battle_portal, # whatever atom you want\n        portal_server: MyApp.Sorcery.PortalServers.Postgres, # See the PortalServer guide\n        query_module: MyApp.Sorcery.Queries.GetBattle, # See the Queries guide\n        query_args: %{player_id: 1} # this will be passed into the query eventually\n      })\n\n    {:ok, socket}\n  end\n\n\n  def handle_event(\"change_hp\", %{\"id\" => idstr, \"amount\" => amountstr}, socket) do\n    id = String.to_integer(idstr)\n    amount = String.to_integer(amountstr)\n\n    # Every mutation requires a portal\n    Sorcery.Mutation.init(socket.assigns.sorcery, :battle_portal)\n    |> Sorcery.Mutation.update([:player, id, :health], fn _old_health, health -> health + amount end)\n    |> Sorcery.Mutation.send_mutation()\n        \n    {:noreply, socket}\n  end\n\n\n  def render(assigns) do\n    ~H\"\"\"\n      <%= for %{id: id, health: health, name: name} = _player <- portal_view(@sorcery, :battle_portal, \"?all_players\") do %>\n         <%= id %> | <%= name %>'s health: <%= health %> \n         Heal  \n         Harm  \n\n      <% end %>\n    \"\"\"\n  end\n\n\nend\n```","title":"LiveViews - live_views","ref":"live_views.html#liveviews"},{"type":"extras","doc":"# Configuration\n\nHerein lies the key to writing a schema. Here is how we would rewrite the Example given by Ecto.Schema:\n\n```elixir\ndefmodule App.User do\n#######\n# We don't need this anymore\n#######\n#\n#use Ecto.Schema\n#\n#schema \"users\" do\n#  field :name, :string\n#  field :age, :integer, default: 0\n#  field :password, :string, redact: true\n#  has_many :posts, Post\n#end\n\n\n  use Sorcery.Schema, %{\n    meta: %{},\n    fields: %{\n        name:      %{t: :string},\n        age:       %{t: :integer, default: 0},\n        password:  %{t: :string,  redact: true},\n        # No need for a has_many. In Sorcery, it is always the child that references the parent.\n    }\n  }\n\n\nend\n```","title":"Configuration","ref":"configuration-4.html"},{"type":"extras","doc":"Before getting into the meat of it, notice the empty map called :meta up there? We can use it to set some schema-wide configuration.\n\n| :meta         | types           | Default       | Description      |\n| ---           | ---             | ---           | ---              |\n| :optional?    | boolean         | true          | Whether every field allows nil. Individual fields can override |\n| :tk           | atom            | :module_name  | The snake case, atom form, of this schemas name. Used in queries etc. |","title":"Meta Map - Configuration","ref":"configuration-4.html#meta-map"},{"type":"extras","doc":"| t: :string    | types           | Optional? | Default | Description      |\n| ---           | ---             | ---       | ---     | ---              |\n| :min          | integer, nil    | true      | nil     | field must have at least this many characters.     |\n| :max          | integer, nil    | true      | nil     | field must have no more than this many characters. |\n| :default      | string, nil     | true      | \"\"      | If no string is given, this will be used.          |\n| :optional?    | boolean         | true      | true    | Whether field is allowed to be nil      |","title":"String fields - Configuration","ref":"configuration-4.html#string-fields"},{"type":"extras","doc":"```elixir\niex> alias Sorcery.Schema.Field.String, as: S\n\n# This fails because \"hi\" is not long enough\niex> S.new(%{min: 5}, \"hi\")\n%S{valid?: false, value: \"hi\", fails: [ {:min, :value} ], ...}\n\n# This passes\niex> S.new(%{min: 5}, \"Hello world, how are you?\")\n%S{valid?: true, value: \"Hello world, how are you?\", fails: [], ...}\n\n# This failse because the schema itself is invalid. :min must be an integer, not a float!\niex> S.new(%{min: 5.6}, \"Hello world, how are you?\")\n%S{valid?: false, value: \"Hello world, how are you?\", fails: [ {:min, :type} ], ...}\n\n# Now we check :max\niex> S.new(%{max: 5}, \"Hello world, how are you?\")\n%S{valid?: false, value: \"Hello world, how are you?\", fails: [ {:max, :value} ], ...}\n\n# By default, strings are optional, with a default of an empty string.\niex> S.new(%{}, nil)\n%S{valid?: true, value: \"\", fails: [], ...}\n\n# ... but you can change that\niex> S.new(%{default: \"Elixir is fun\"}, nil)\n%S{valid?: true, value: \"Elixir is fun\", ...}\n\niex> S.new(%{optional?: false}, nil)\n%S{valid?: false, value: nil, fails: [ {:optional?, :value} ] }\n\n# Note that optional?: false does nothing if a default is set.\niex> S.new(%{optional?: false, default: \"Elixir is fun\"}, nil)\n%S{valid?: true, value: \"Elixir is fun\", ...}\n\n```","title":"String Examples - Configuration","ref":"configuration-4.html#string-examples"},{"type":"extras","doc":"|               | types           | Optional? | Default | Description                     |\n| ---           | ---             | ---       | ---     | ---                             |\n| :min          | integer, nil    | true      | nil     | field be at least this big.     |\n| :max          | integer, nil    | true      | nil     | field be no more than this big. |\n| :default      | integer, nil    | true      | nil     | If no value is given, this will be used.          |\n| :optional?    | boolean         | true      | true    | If no value is given, this field is invalid.      |","title":"Integer fields - Configuration","ref":"configuration-4.html#integer-fields"},{"type":"extras","doc":"```elixir\niex> alias Sorcery.Schema.Field.I, as: I\n\n# This fails because 2 is not big enough\niex> I.new(%{min: 5}, 2)\n%S{valid?: false, value: 2, fails: [ {:min, :value} ], ...}\n\niex> I.new(%{min: 5}, 1337)\n%S{valid?: true, value: 1337, fails: [], ...}\n\niex> I.new(%{max: 5}, 2)\n%S{valid?: true, value: 2, fails: [], ...}\n\niex> I.new(%{}, nil)\n%S{valid?: true, value: nil, fails: [], ...}\n\niex> I.new(%{default: 5}, nil)\n%S{valid?: true, value: 5, fails: [], ...}\n\niex> I.new(%{optional?: false}, nil)\n%S{valid?: false, value: nil, fails: [ {:optional?, :value} ], ...}\n\niex> I.new(%{optional?: false, default: 5}, nil)\n%S{valid?: true, value: 5, fails: [], ...}\n```","title":"Integer Examples - Configuration","ref":"configuration-4.html#integer-examples"},{"type":"extras","doc":"# Introduction\n\nSorcery Query Language is a reversible query language. Please don't call it SQL, that would be confusing for some reason. SrcQL will do fine.\n\nThe syntax is partially inspired by Datalog, which is used by the Datomic database. But there are many important differences, and it is not intended to be identical.","title":"Introduction","ref":"introduction-2.html"},{"type":"extras","doc":"As a quirky example, we are making a multiplayer game in which there are players, organized into teams, and each player is a wizard who casts spells. They can all battle in an 'arena'\nSo the four tables are: team, player, spell, arena\n\nNow suppose we want a query that pulls up data for arena 1 \n\n```elixir\n# Don't try to fully understand this yet. We're going to break it down soon and explain it line by line.\n\nalias Sorcery.Query, as: SrcQL\nSrcQL.new(%{\n  where: [\n    [\"?arena\",   :arena,   :id, :args_arena_id],\n    [\"?teams\",   :team,    :current_arena_id, \"?arena.id\"],\n    [\"?players\", :player, [\n      {:team_id, \"?teams.id\"},\n      {:health, {:>, 0}}\n    ]],\n    [\"?spells\",  :spell,   :player_id, \"?players.id\"]\n  ],\n  args: %{\n    arena_id: :integer\n  },\n  find: %{\n    \"?players\" => :*,\n    \"?spells\" =>  [:name, :energy, :type, :player_id]\n  }\n})\n\nargs = %{arena_id: 1}\n\n\n# Now when you assign this Query to a Portal called :portal_name, and you pass in the args above, you can forevermore view the \"?players\" and \"?spells\" lvars\niex> portal_view(@sorcery, :portal_name, \"?players\")\n[\n  %{id: 2, name: \"Jose\", health: 100, team_id: 1},\n  %{id: 85, name: \"Aaron\", health: 10, team_id: 2},\n]\n```\n\n\nSo you might have guessed the list of :where tuples is similar to something you might find in SQL. Just a list of conditions for filtering data. Except the tuples look weeeeiiird.\n\n```elixir\n  # each tuple has 4 elements in it.\n  # {lvar,     tk,       attr,  value}\n  {\"?arena\",   :arena,   :id,   1}, \n```\nThe string `\"?arena\"` is a logic variable, or Lvar for short.\nit represents a lazy Set of entities of a specific type. in this case they are all :arena entities.\nThe attr/value pair helps us limit the set of matches. Now every entity in the ?arena set must have an :id of 1. \nSo how big is the set going to be? Either 0 or 1.\n\nLet's continue.\n\n```elixir\n  # Now the final element (the value) is referencing a previous lvar... and appending \".id\" to the end of it.\n  # If you imagine gettings all the possible ids from ?arena, i.e. ids = Enum.map(?arena, &(&1.id))\n  # And then we are basically running Enum.filter(?teams, &(&1.current_arena_id in ids))\n  {\"?teams\",   :team,    :current_arena_id, \"?arena.id\"}\n```\nAh yes, our first inferred join. Instead of a JOIN...ON syntax, we just do this and it works like ~~magic~~ sorcery.\n\n\n```elixir\n  # Boring, you know this one already.\n  {\"?players\", :player,  :team_id, \"?teams.id\"}\n\n  # Ahh now we bring operators into the mix!\n  # So we only get players who are members of at least one of the ?teams...\n  # But they ALSO must not be dead. Makes it hard to play. So it filters player.health > 0\n  {\"?players\", :player,  :health, {:>, 0}}\n```\n\nBut what about :find? Well it's kind of like if a SQL select statement made a baby with Map.take/2.\nNote that it automatically adds an :id field to every take... and the format returned is a map in a pretty specific shape.\n```elixir\nfind: %{\"?foo\" => [:name] }\n\n# lvar => %{id => entity}\n\nselection = %{\n    \"?foo\" => %{\n        1 => %{id: 1, name: \"some name\"}\n    }\n}\n```\n\n@TODO have the option to exclude :id fields.\n\n@TODO have the option to use a different primary key than :id.","title":"Basic Usage - Introduction","ref":"introduction-2.html#basic-usage"},{"type":"extras","doc":"What if I told you the :where tuples were actually just a shorthand for a more detailed syntax?\n\nLet's rewrite the query\n\n\n```elixir\n\nalias Sorcery.Query, as: SrcQL\nSrcQL.new(%{\n  where: [\n    SrcQL.Clause.new(%{lvar: \"?arena\", tk: :arena, filters: [%{attr: :id, val: 1, op: :==}]}),\n    SrcQL.Clause.new(%{lvar: \"?teams\", tk: :team,  filters: [%{attr: :team_id, lvar: \"?arena\", lvar_attr: \"id\", op: :in}] }),\n    SrcQL.Clause.new(%{lvar: \"?players\", tk: :player, filters: [\n      %{attr: :team_id, lvar: \"?arena\", lvar_attr: \"id\", op: :in},\n      %{attr: :health, val: 0, op: :>}\n    ]}),\n    SrcQL.Clause.new(%{lvar: \"?spells\", tk: :spell, filters: [%{attr: :player_id, lvar: \"?players\", lvar_attr: \"id\", op: :in}]}),\n  ],\n  find: %{\n    \"?players\" => [:name, :health, :team_id]\n    \"?spells\" =>  [:name, :energy, :type, :player_id]\n  }\n})\n```\nThese functions return structs %SrcQL.Clause{...}\nAs do the tuples. It's all the same.\n\nNow why on earth would you want this? Well you probably don't want to do it manually. But it's useful to know, because there are also other structs available, for different types of clauses!\n\n```elixir\n# This will get all foo entities with :magic greater than 100, OR with an id of 42.\nSrcQL.OrClause.new(%{lvar: \"?foo\", tk: :foo, filters: [\n  %{attr: :id, val: 42},\n  %{attr: :magic, val: 100, op: :>},\n]})\n```\nThere is no need for an AndClause, because the default Clause does that.\n\nYou can also do certain aggregation operations.\n\n```elixir\n# This will get the average health of all entities matching \"?players\"\nSrcQL.Aggregate.Mean.new(%{lvar: \"?avg_health\", val: \"?players.health\"})\n```","title":"Function syntax - Introduction","ref":"introduction-2.html#function-syntax"},{"type":"extras","doc":"# Introduction\n\nPortals are what tie it all together. The connective tissue of Sorcery. But first we have one other concept to talk about.","title":"Introduction","ref":"introduction-1.html"},{"type":"extras","doc":"A PortalServer is an elixir process which has access to a data store. When you send a SorceryQuery to a PortalServer, you don't *just* get some data back... you create a Portal, with a continuous feed of data matching that query. Kind of like a PubSub on steroids.\n\nThe PortalServer holds all the configuration, describing in detail HOW the data is accessed. \n\nHere you can throw in your own custom adapters for different databases. \n\nYou can tell it to pass messages via REST and GraphQL or elixir's Process.send\n\nAs long as you find/build the right adapters, the options are limitless.","title":"PortalServers - Introduction","ref":"introduction-1.html#portalservers"},{"type":"extras","doc":"| PortalServer      | Data Store                  |\n| ---               |  ---                        |\n| Phoenix.LiveView  |  assigns.sorcery            |\n| GenServer         |  a map in the state         |\n| GenServer         |  Repo connected to Postgres |\n| GenServer         |  ANOTHER PortalServer       |\n| A plain Process   |  An external weather API    |\n\nDo you see how flexible it is? Very. I believe one could technically have a PortalServer fire off an email, which someone manually responds to with an attachment... as long as you write an adapter for it, and don't care about latency, it would work.","title":"Fun examples of PortalServers to get your brain juices flowing: - Introduction","ref":"introduction-1.html#fun-examples-of-portalservers-to-get-your-brain-juices-flowing"},{"type":"extras","doc":"This gets special mention because it was the original use case, and reason for creating Sorcery.\n\n```elixir\n  def mount(_, _, socket) do\n    socket = socket\n             |> Sorcery.PortalServer.create()\n             |> Sorcery.Portal.open(to, sorcery_query)\n    {:ok, socket}\n  end\n\n  def handle_info({:sorcery, msg}, socket), do: Sorcery.PortalServer.live_view_handler(msg, socket)\n```\n\nThat is all we need to do to turn a LiveView into a PortalServer. And just for fun, we also open up a Portal to some other PortalServer. See the guides on Queries and LiveViews for more detail.\n\nNote that there will be times when this LiveView receives all sorts of whacky messages from whatever other PortalServers it connects to. You don't need to care about those whatsoever, they will all be tagged as {:sorcery, _}\n\nNow you can obviously intercept those and look at them with IO.inspect, or dbg, just for fun. But you do not need to know what's going on behind the curtain. I am a powerful wizard, you can trust me.","title":"LiveView - Introduction","ref":"introduction-1.html#liveview"},{"type":"extras","doc":"# Configuration\n\nThe cheatsheet for building a PortalServer\n\n```elixir\ndefmodule Src.PortalServers.Demo do\n  use GenServer\n\n  @config %{\n    store_args: %{repo: MyApp.Repo},\n    store_plugin: Sorcery.Plugins.Ecto\n  }\n\n  def init() do\n  end\n\n\nend\n```\n\n\n| t: :string    | types           | Optional? | Default                    | Description      |\n| ---           | ---             | ---       | ---                        | ---              |\n| :prune_dead   | integer         | false     | 60_000                     | How often (ms) do we remove queries from dead pids. |\n| :store_plugin | module          | false     | Sorcery.Plugins.GenServer  | Module for reading and writing to storage           |\n| :store_args   | map             | false     | %{}                        | Any extra data needed by the store_plugin goes here |","title":"Configuration","ref":"configuration-3.html"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,51,46,48]}}